import os
import types

from etils import epath, epy

from apitree import structs, symbol_match, tree_extractor


def write_doc(
    info: structs.ModuleInfo,
    *,
    verbose=True,
    root_dir: epath.Path = None,
) -> None:
  node = tree_extractor.get_api_tree(info)
  if not root_dir:
    root_dir = epath.resource_path(node.symbol.value)
    root_dir = root_dir.parent / 'docs/api'

  if verbose:
    print(node)

  _write_node(root_dir, node)


def _write_node(root_dir: epath.Path, node: tree_extractor.Node) -> None:
  file = root_dir / node.match.filename
  file.parent.mkdir(exist_ok=True, parents=True)
  file.write_text(epy.dedent(node.match.content))

  for child in node.documented_childs:
    _write_node(root_dir, child)



from __future__ import annotations

import ast
import contextlib
import dataclasses
import functools
import inspect
import os
import types

from etils import epy

from apitree import import_utils


@dataclasses.dataclass
class ImportAlias:
  """Represents an import symbol."""

  namespace: str
  alias: str


class _GlobalImportVisitor(ast.NodeVisitor):

  def __init__(self):
    self.symbols = []

  def visit_Import(self, node):
    for alias in node.names:
      self.symbols.append(
          ImportAlias(alias.name, alias.asname or alias.name.split('.', 1)[0])
      )
    self.generic_visit(node)

  def visit_ImportFrom(self, node):
    module = node.module or ''
    for alias in node.names:
      self.symbols.append(
          ImportAlias(f'{module}.{alias.name}', alias.asname or alias.name)
      )
    self.generic_visit(node)

  def generic_visit(self, node):
    if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
      return
    return super().generic_visit(node)


def parse_global_imports(content: str | types.ModuleType) -> list[ImportAlias]:
  """Extracts import symbols from a Python module.

  Args:
    module: The Python module to extract import symbols from.

  Returns:
    A list of Symbol objects representing the import symbols.

  """
  if isinstance(content, types.ModuleType):
    content = inspect.getsource(content)
  tree = ast.parse(content)

  visitor = _GlobalImportVisitor()
  visitor.visit(tree)

  return visitor.symbols


@dataclasses.dataclass
class _SymbolDefinition:
  module_name: str
  start: int
  end: int

  @property
  def git_lno(self) -> str:
    if self.start == self.end:
      return f'#L{self.start}'
    else:
      return f'#L{self.start}-L{self.end}'

  @property
  def belong_to_project(self) -> bool:
    return import_utils.belong_to_repo(self.module_name)

  @property
  def filename(self) -> str:
    return os.fspath(import_utils.repo_relative_path(self.module_name))

  @property
  def last_project_symbol(self) -> _SymbolDefinition:
    if not self.belong_to_project:
      raise ValueError(
          f'{self.module_name} is not part of the project.\n'
          f' * Repo: {import_utils.repo_path()}\n'
          f' * Proj: {import_utils.abs_path(self.module_name)}\n'
          'Make sure that the project was installed locally (`pip install'
          ' -e .`)'
      )
    return self

  @property
  def docstring(self) -> str:
    # TODO(epot): Should also support `Attributes:` docstring
    # What if the `Attributes:` is hidden in the import chain ?
    lines = import_utils.module_lines(self.module_name)
    docstring_lines = []
    for lines_no in range(self.start - 2, 0, -1):
      line = lines[lines_no].strip()
      if not line.startswith('# '):
        break
      line = line.removeprefix('# ')
      docstring_lines.append(line)
    return '\n'.join(reversed(docstring_lines))

  @property
  def code(self) -> str:
    lines = import_utils.module_lines(self.module_name)
    lines = lines[self.start - 1 : self.end]
    return '\n'.join(lines)


@dataclasses.dataclass
class _ImportedSymbol(_SymbolDefinition):
  import_module_name: str
  symbol_name: str

  @property
  def last_project_symbol(self) -> _SymbolDefinition:
    if not self.belong_to_project:
      raise ValueError(f'{self.module_name} is not part of the project.')

    # No need to load the child if it do not belong to the module
    if import_utils.repo_relative_path(self.import_module_name) is None:
      return self

    sub_symbols = extract_symbols(self.import_module_name)
    if self.symbol_name not in sub_symbols:
      raise ValueError(
          f'`{self.symbol_name}` not found in `{self.import_module_name}`'
      )
    sub = sub_symbols[self.symbol_name]
    if not sub.belong_to_project:  # The child do not bellong to the project
      return self

    return sub.last_project_symbol  # Recurse


@dataclasses.dataclass
class _AssignedSymbol(_SymbolDefinition):
  pass


class _DeclaredSymbol(_SymbolDefinition):
  pass


class _GlobalAssignementExtractor(ast.NodeVisitor):

  def __init__(self, module_name: str):
    self._module_name = module_name
    self.symbols: dict[str, _SymbolDefinition] = {}

    self._should_capture = None

  @contextlib.contextmanager
  def _capture(self, node: ast.AST):
    assert not self._should_capture
    self._should_capture = _AssignedSymbol(
        module_name=self._module_name,
        start=node.lineno,
        end=node.end_lineno,
    )
    yield
    assert self._should_capture
    self._should_capture = None

  # TODO(epot): Support `ast.NamedExpr` (`:=`)

  def visit_Name(self, node: ast.Name):  # x = y
    if self._should_capture:
      self.symbols[node.id] = self._should_capture

  def visit_Assign(self, node: ast.Assign):  # x = y
    with self._capture(node):
      for target in node.targets:
        self.visit(target)

  def visit_AnnAssign(self, node: ast.AnnAssign):  # x: int = y
    with self._capture(node):
      self.visit(node.target)

  # `import xxx as yyy` only import modules

  def visit_ImportFrom(self, node):
    module = node.module or ''
    for alias in node.names:
      alias_name = alias.asname or alias.name
      self.symbols[alias_name] = _ImportedSymbol(
          module_name=self._module_name,
          start=node.lineno,
          end=node.end_lineno,
          import_module_name=module,
          symbol_name=alias.name,
      )
    self.generic_visit(node)

  def generic_visit(self, node):
    if isinstance(
        node,
        (
            ast.FunctionDef,
            ast.AsyncFunctionDef,
            ast.ClassDef,
            ast.Attribute,  # x.y
            ast.Subscript,  # x[0]
        ),
    ):
      # Skip recursing inside functions,...
      return
    return super().generic_visit(node)


@functools.cache
def extract_symbols(module_name: str) -> dict[str, _SymbolDefinition]:
  path = import_utils.module_path(module_name)
  return _extract_assignement_lines(path.read_text(), module_name)


@functools.cache
def extract_last_symbol(
    module_name: str, name: str
) -> _SymbolDefinition | None:
  try:
    symbols = extract_symbols(module_name)
  except Exception as e:
    epy.reraise(e, prefix=f'{module_name}:{name}: ')
  if name not in symbols:
    return
  symbol = symbols[name]
  symbol = symbol.last_project_symbol
  return symbol


def _extract_assignement_lines(
    file_content: str,
    module_name: str,
) -> dict[str, _SymbolDefinition]:
  tree = ast.parse(file_content)

  extractor = _GlobalAssignementExtractor(module_name=module_name)
  extractor.visit(tree)
  return extractor.symbols



"""Add links to GitHub source code."""

import functools
import importlib
import inspect
import os
import pathlib
import subprocess

from apitree import ast_utils, debug_utils, import_utils


@debug_utils.print_error()
def linkcode_resolve(domain, info):
  if domain != 'py':
    return None
  if not info['module']:
    return None
  return _linkcode_resolve(info['module'], info['fullname'])


def _linkcode_resolve(module_name: str, fullname: str) -> str:
  suffix = _get_lines_suffix(module_name, fullname)
  if not suffix:
    if '.' in fullname:
      # Attributes are not documented (as they beloong to different files)
      return None
    # Fallback on the module name.
    symbol = ast_utils.extract_last_symbol(module_name, fullname)
    if symbol is None:
      raise ValueError(f'{module_name}:{fullname} not found.')
    suffix = f'{symbol.filename}{symbol.git_lno}'
  return f'{_get_github_url()}/tree/main/{suffix}'


def get_module_link(module_name):
  path = import_utils.repo_relative_path(module_name)
  return f'{_get_github_url()}/tree/main/{path}'


def get_assignement_link(module_name, name):
  symbol = ast_utils.extract_last_symbol(module_name, name)
  if symbol is None:
    raise ValueError(f'{module_name}:{name} not found.')

  source_link = (
      f'{_get_github_url()}/tree/main/{symbol.filename}{symbol.git_lno}'
  )
  return source_link


def _get_lines_suffix(module_name: str, qualname: str) -> str:
  module = importlib.import_module(module_name)

  obj = module
  for part in qualname.split('.'):
    try:
      obj = getattr(obj, part)
    except AttributeError:
      return ''  # Object unavailable

  # Unwrap until the last file inside the project
  try:
    obj = _unwraps(obj)
  except _NotBelongToModule:
    # For example:
    # `@functools.wraps(external_fn)(fn)`
    # `from external_module import fn`
    return os.fspath(import_utils.repo_relative_path(module_name))

  # After unwrapping, the object might be in a different file
  module_name = _get_module_name(obj)
  if module_name is None:
    return ''

  try:
    lines = inspect.getsourcelines(obj)
  except (TypeError, OSError):
    return ''

  # Detect if file is inside the project (and not `Union` or similar)
  filepath = import_utils.repo_relative_path(module_name)
  if not filepath:
    return ''

  start = lines[1]
  end = start + len(lines[0]) - 1
  return f'{filepath}#L{start}-L{end}'


@functools.cache
def _get_github_url() -> str:
  # TODO(epot): Support cross-repo
  out = subprocess.run(
      'git config --get remote.origin.url',
      shell=True,
      capture_output=True,
      text=True,
  )
  url = out.stdout.strip()
  assert url.startswith('https://github.com')
  return url


# Could try to use `app.builder.env` to auto-setup the extension


def _unwraps(fn):
  if not hasattr(fn, '__wrapped__'):  # Not a function, decorator
    return fn

  # Note: Do not support cycles.
  chain = [fn]
  while hasattr(fn, '__wrapped__'):
    fn = fn.__wrapped__
    chain.append(fn)

  for fn in reversed(chain):
    module_name = _get_module_name(fn)
    if module_name is None:  # Class attributes (e.g. `x: int`)
      continue
    if import_utils.belong_to_repo(module_name):
      return fn
  # Symbol belong to another module
  raise _NotBelongToModule(f'Not bellong to repo for symbol {fn}')


def _get_module_name(obj) -> str:
  try:
    _module = inspect.getmodule(obj)
    if _module is None:  # Class attributes (e.g. `x: int`)
      return None
      # raise ValueError(f'Not found: {module_name} {qualname}')
    module_name = _module.__name__
  except TypeError:
    return None
  return module_name


class _NotBelongToModule(Exception):
  pass



import functools
import os
import pathlib
import sys
import tomllib
from typing import Any

import sphinx
from etils import epath, epy

from apitree import import_utils, structs, writer
from apitree.ext import github_link


def setup(app, *, callbacks):
  for callback in callbacks:
    callback()


def make_project(
    *,
    modules: dict[str, str] | structs.ModuleInfo,
    includes_paths: dict[str, str] = {},
    globals: dict[str, Any],
) -> None:
  """Setup the `conf.py`.

  Args:
    modules: Top module names to document.
    includes_paths: Mapping to external files to `docs/` path (e.g.
      `my_module/submodule/README.md` to `submodule.md`). By default, only
      files inside `docs/...` can be read
    globals: The `conf.py` `globals()` dict. Will be mutated.
  """

  docs_dir = epath.Path(globals['__file__']).parent  # <repo>/docs/
  repo_dir = docs_dir.parent
  # TODO(epot): Fragile if one of the module is already imported.
  # If so, should check that imported modules are
  # `import_utils.belong_to_project`

  # Clear etils before (to allow documenting `etils` itself).
  # Otherwise, `etils` imported is the one from `pip install` and
  # not the one from `git clone`.
  _clear_etils()
  sys.path.insert(0, os.fspath(repo_dir))

  project_name = _get_project_name(repo_dir=repo_dir)

  # API generator
  api_ext = 'sphinx.ext.autodoc'
  api_ext_config = dict(
      autodoc_typehints_format='fully-qualified',  # `x.y.MyClass`
      autodoc_default_options={
          'members': True,
          'show-inheritance': True,
          'member-order': 'bysource',
          'undoc-members': True,
      },
  )

  # Uncomment to try autoapi
  # api_ext = 'autoapi.extension'
  # api_ext_config = dict(
  #     autoapi_dirs=[f'../{project_name}'],
  #     autoapi_ignore=['*migrations*', '*_test.py'],
  #     autoapi_keep_files=True,
  #     autoapi_python_use_implicit_namespaces=True,
  # )

  globals.update(
      # Project information
      project=project_name,
      copyright=f'2023, {project_name} authors',
      author=f'{project_name} authors',
      # General configuration
      extensions=[
          api_ext,  # API Doc generator
          'myst_nb',  # Notebook support
          'sphinx.ext.napoleon',  # Numpy-style docstrings
          # 'sphinx.ext.linkcode',  # Links to GitHub
          # Others:
          # 'sphinx_autodoc_typehints',
          # 'sphinx.ext.linkcode',
          # 'sphinx.ext.inheritance_diagram',
          # 'myst_parser',
          # API Tree
          'apitree.ext.docstring',  # Fix bad ```python md formatting
          'apitree.ext.auto_ref',  # Add cross ref for inline code
      ],
      exclude_patterns=[
          '_build',
          'jupyter_execute',
          'Thumbs.db',
          '.DS_Store',
      ],
      # HTML output
      html_theme='sphinx_book_theme',
      # Other themes:
      # 'alabaster' (default)
      # 'sphinx_material'
      # 'sphinx_book_theme' (used by Jax)
      html_title=project_name,
      # TODO(epot): Instead should have a self-reference TOC
      html_theme_options={'home_page_in_toc': True},
      # -- Extensions ---------------------------------------------------
      # ---- myst -------------------------------------------------
      myst_heading_anchors=3,
      # ---- myst_nb -------------------------------------------------
      nb_execution_mode='off',
      # ---- api extension -------------------------------------------------
      **api_ext_config,
      # Register hooks
      setup=functools.partial(
          setup,
          callbacks=[
              functools.partial(
                  _write_api_doc, docs_dir=docs_dir, modules=modules
              ),
              functools.partial(
                  _write_include_paths,
                  repo_dir=repo_dir,
                  docs_dir=docs_dir,
                  includes_paths=includes_paths,
              ),
          ],
      ),
      # ---- linkcode -------------------------------------------------
      linkcode_resolve=github_link.linkcode_resolve,
  )


def _write_api_doc(
    *,
    docs_dir: pathlib.Path,
    modules: dict[str, str] | structs.ModuleInfo,
):
  api_dir = docs_dir / 'api'
  if api_dir.exists():
    api_dir.rmtree()
  api_dir.mkdir()

  if isinstance(modules, dict):
    modules = [
        structs.ModuleInfo(alias=k, module_name=v) for k, v in modules.items()
    ]
  if isinstance(modules, structs.ModuleInfo):
    modules = [modules]

  for module_info in modules:
    writer.write_doc(module_info, root_dir=api_dir)


def _write_include_paths(
    *,
    repo_dir: pathlib.Path,
    docs_dir: pathlib.Path,
    includes_paths: dict[str, str],
):
  for repo_path, doc_path in includes_paths.items():
    src_path = repo_dir / repo_path
    dst_path = docs_dir / doc_path
    match src_path.suffix:
      case '.md':
        # repo_dir.parent / 'etils'
        # Could dynamically compute the `../../../`
        content = epy.dedent(
            f"""
            ```{{include}} ../{repo_path}
            ```
            """
        )
        dst_path.write_text(content)
      case '.ipynb':
        dst_path.write_bytes(src_path.read_bytes())
      case default:
        raise ValueError(f'Invalid suffix: {default}')


def _get_project_name(repo_dir):
  # TODO(epot): This hardcode too much assumption on the program
  path = repo_dir / 'pyproject.toml'
  info = tomllib.loads(path.read_text())
  return info['project']['name']


def _clear_etils():
  # For re-triggering etils import (as it is used both for documentation
  # and in apitree
  for module_name in list(sys.modules):
    if module_name.startswith('etils'):
      del sys.modules[module_name]



from __future__ import annotations

import dataclasses
import functools
import importlib
import types
from typing import Iterator

from etils import edc, epy

from apitree import context, structs, symbol_match


@dataclasses.dataclass
class Node:
  symbol: symbol_match.Symbol

  def __post_init__(self):
    self.symbol.node = self
    context.add_ref(self)

  @property
  def match(self):
    return self.symbol.match

  @functools.cached_property
  def childs(self) -> list[Node]:
    if not self.match.recurse:
      return []
    module = self.symbol.value
    all_childs = []
    for k, v in module.__dict__.items():
      symbol = symbol_match.Symbol(
          name=k,
          value=v,
          parent=module,
          parent_symb=self,
          ctx=self.symbol.ctx,
      )
      all_childs.append(Node(symbol))

    return all_childs

  @functools.cached_property
  def documented_childs(self) -> list[Node]:
    return [n for n in self.childs if n.match.documented]

  def iter_documented_nodes(self) -> Iterator[Node]:
    yield self
    for c in self.documented_childs:
      yield from c.iter_documented_nodes()

  def __repr__(self) -> str:
    if self.childs:
      child_str = epy.Lines.repr([c for c in self.childs if c.match.documented])
    else:
      child_str = ''
    return f'{self.symbol.name}={type(self.match).__name__}{child_str}'


def get_api_tree(info: structs.ModuleInfo):
  module = importlib.import_module(info.api)

  # ctx = symbol_match.Context(
  #     module_name=module.__name__,
  #     info,
  # )
  return Node(
      symbol_match.Symbol(
          name=module.__name__,
          value=module,
          parent=None,
          parent_symb=None,
          ctx=info,
      ),
  )



"""Automatically add cross-links in markdown.

Replace all `my_module.XXX` by `:ref:...` to link to the API doc.

"""
import os
import pathlib
import typing

from docutils import nodes
from sphinx.application import Sphinx

from apitree import context, debug_utils


def _is_inside_link(node: nodes.Node):
    while node.parent:
        if isinstance(node.parent, nodes.reference):
            return True
        node = node.parent
    return False


def _add_refs(app: Sphinx, doctree: nodes.document, docname: str):
    docpath = pathlib.Path(docname + '.md')

    for node in doctree.findall(lambda n: isinstance(n, (nodes.title_reference, nodes.literal))):
        node = typing.cast(nodes.Element, node)

        if _is_inside_link(node):
            continue

        ref_name = node.astext()
        ref_uri = context.get_ref(ref_name.lstrip('@'))

        if ref_uri is None:
            continue

        ref_path = pathlib.Path('api') / ref_uri.match.filename
        ref_path = ref_path.with_suffix('.html')

        ref_path = _relative_path(ref_path, docpath)

        # Wrap inside ref
        ref = nodes.reference(refuri=os.fspath(ref_path))
        ref += nodes.literal(text=ref_name)

        node.replace_self(ref)


def _relative_path(p0: pathlib.Path, p1: pathlib.Path) -> pathlib.Path:
  if p0.is_relative_to(p1):
    return p0.relative_to(p1)

  # Extract the common part between the paths
  num_common_parts = 0
  for part0, part1 in zip(p0.parts, p1.parts):
    if part0 != part1:
      break
    num_common_parts +=1

  num_parents = len(p1.parts) - num_common_parts -1

  new_p = pathlib.Path(*(['..'] * num_parents)).joinpath(*p0.parts[num_common_parts:])
  return new_p



def setup(app: Sphinx):
    app.connect('doctree-resolved', debug_utils.print_error()(_add_refs))



"""."""

import dataclasses

from etils import edc


@edc.dataclass
@dataclasses.dataclass
class ModuleInfo:
  """.

  Attributes:
    api: Entry point of the API
    module_name: What to include
    alias: Short name of the module
  """

  module_name: str
  api: str = None
  alias: str = None

  def __post_init__(self):
    if self.api is None:
      self.api = self.module_name
    if self.alias is None:
      self.alias = self.module_name



from __future__ import annotations

import collections
import os
from typing import Optional

from apitree import tree_extractor


class Context:

  def __init__(self):
    self.refs: dict[str, list[tree_extractor.Node]] = collections.defaultdict(list)


def get_ref(name: str) -> Optional[tree_extractor.Node]:
  matches = _context.refs.get(name)
  if not matches:
    return None
  if len(matches) != 1:
    return None
  return  matches[0]


def add_ref(node: tree_extractor.Node) -> None:
  name = node.symbol.qualname
  _context.refs[name].append(node)
  if '.' in name:
    _, subname = name.rsplit('.', 1)
    _context.refs[subname].append(node)
  # `qualname_no_alias` is acually the real module name.
  # Could also add the real `qualname_no_alias`. For example:
  # * `kd.typing.XX`  # Public symbol
  # * `kauldron.typing.XX`  # Alias replaced
  # * `kauldron.utils.typing.XX`  # Real module location
  if node.symbol.qualname_no_alias != node.symbol.qualname:
    _context.refs[node.symbol.qualname_no_alias].append(node)



_context = Context()



from __future__ import annotations

import dataclasses
import enum
import functools
import os
import pathlib
import types
import typing
from collections.abc import Callable, Iterator
from typing import Any

import typing_extensions
from etils import edc, epath, epy

from apitree import ast_utils, md_utils, tree_extractor
from apitree.ext import github_link


@edc.dataclass
@dataclasses.dataclass
class Context:
  module_name: str
  alias: str
  # visited: dict[str, _RefNode] = dataclasses.field(default_factory=dict)


class SymbolType(enum.StrEnum):
  MODULE = enum.auto()
  CLASS = enum.auto()
  FUNCTION = enum.auto()
  ATTRIBUTE = enum.auto()
  TYPING = enum.auto()
  UNKNOWN = enum.auto()


@edc.dataclass
@dataclasses.dataclass
class Symbol:
  name: str
  value: Any

  parent: types.ModuleType
  parent_symb: tree_extractor.Node
  ctx: Context = dataclasses.field(repr=False)

  node: tree_extractor.Node = dataclasses.field(repr=False, init=False)

  @functools.cached_property
  def is_imported(self) -> bool:
    return self.name in self.imported_symbols

  @functools.cached_property
  def imported_symbols(self) -> set[str]:
    if self.parent.__file__ is None:  # Implicit package
      return set()
    # TODO(epot): Cache across module
    return set(imp.alias for imp in ast_utils.parse_global_imports(self.parent))

  @functools.cached_property
  def belong_to_namespace(self) -> bool:
    return self.value.__name__.startswith(self.ctx.module_name)

  @functools.cached_property
  def match(self) -> type[Match]:
    return Match.root_match(self)

  @functools.cached_property
  def qualname(self) -> str:
    """Exemple: `dca.typing.Float`."""
    if self.parent_symb is None:  # root node
      # assert self.ctx.module_name == self.name
      return self.ctx.alias
    return f'{self.parent_symb.symbol.qualname}.{self.name}'

  @functools.cached_property
  def qualname_no_alias(self) -> str:
    """Exemple: `dataclass_array.typing.Float`."""
    if isinstance(self.value, types.ModuleType):
      try:
        return self.value.__name__
      except Exception:  # TODO(epot): Better lazy-modules
        return ''
    else:
      return f'{self.parent_symb.symbol.qualname_no_alias}.{self.name}'

  # Return type


class Match:
  symbol: Symbol

  recurse: bool = False
  documented = True
  template_name: str | None = None
  docstring_1line: str = ''
  icon: SymbolType = SymbolType.UNKNOWN

  extra_template_kwargs = {}

  SUBCLASSES: list[Match] = []
  SKIP_REGISTRATION = False

  def __init__(self, symbol):
    self.symbol = symbol

  def __init_subclass__(cls) -> None:
    if 'SUBCLASSES' not in cls.__dict__:
      cls.SUBCLASSES = []
    for parent_cls in cls.__bases__:
      parent_cls.SUBCLASSES.append(cls)

  @classmethod
  def root_match(cls, symbol: Symbol) -> Match:
    for subcls in cls.SUBCLASSES:
      if subcls.__dict__.get('SKIP_REGISTRATION', False):
        continue
      all_match = []
      self = subcls(symbol)
      for subcls_parents in subcls.mro():
        if subcls_parents is object:
          continue
        if subcls_parents.__dict__.get('SKIP_REGISTRATION', False):
          continue
        try:
          all_match.append(subcls_parents.match(self))
        except Exception as e:
          epy.reraise(
              e, prefix=f'{symbol} ({cls.__name__}) for {subcls.__name__}'
          )

      if all(all_match):  # Found match
        return subcls.root_match(symbol)  # Maybe missing values, recurse
    else:
      if not cls.SUBCLASSES:  # Leaf
        return cls(symbol)
      raise ValueError(f'No match found for {symbol} (in {cls})')

  def match(cls) -> bool:
    return True

  # Write the doc

  @property
  def filename(self) -> pathlib.Path:
    raise ValueError(f'Missing filename for {type(self)}')

  @functools.cached_property
  def template(self) -> str:
    if not self.template_name:
      return ''
    return load_template(self.template_name)

  @property
  def content(self):
    # TODO(epot)
    # * Title
    # * Docstring
    # * Signature
    # * Arguments
    # * Source code
    return self.template.format(
        qualname=self.symbol.qualname,
        qualname_no_alias=self.symbol.qualname_no_alias,
        **self.extra_template_kwargs,
    )

  def make_symbols_table(self, nodes: Iterator[tree_extractor.Node]):
    table = md_utils.Table(header=['', ''])

    for n in nodes:
      filename = n.match.filename
      filename = filename.relative_to(self.filename.parent)
      filename = os.fspath(filename)
      filename = filename.removesuffix('.md')
      table.add_row(
          # f'*{n.match.icon}*',
          f'[{n.symbol.qualname}]({filename})',
          f'{n.match.docstring_1line}',
      )

    return table.make()

  def make_symbols_tables(self, nodes: tree_extractor.Node) -> str:
    type_to_childs = epy.groupby(nodes, key=lambda n: n.match.icon)
    lines = []
    for type_ in SymbolType:
      if type_ not in type_to_childs:
        continue
      lines.append('')
      lines.append(f'### {type_.capitalize()}')
      lines.append('')

      childs = type_to_childs[type_]
      childs = sorted(childs, key=lambda n: n.symbol.qualname)
      lines.append(self.make_symbols_table(childs))
    return '\n'.join(lines)


def _not(cls: type[Match]) -> Callable[[Match], bool]:
  def match(self):
    return not cls.match(self)

  return match


@functools.cache
def load_template(template_name):
  path = epath.resource_path('apitree') / f'templates/{template_name}.md'
  return path.read_text()


class _WithDocstring(Match):
  SKIP_REGISTRATION = True

  @property
  def docstring_1line(self) -> str:
    if not getattr(self.symbol.value, '__doc__', None):
      return ''
    else:
      return self.symbol.value.__doc__.split('\n', 1)[0]


class _IsModule(_WithDocstring, Match):
  icon = SymbolType.MODULE
  template_name = 'module'

  def match(self) -> bool:
    return isinstance(self.symbol.value, types.ModuleType)

  @property
  def filename(self) -> pathlib.Path:
    return (
        self.symbol.parent_symb.match.filename.parent
        / self.symbol.name
        / 'index.md'
    )

  @property
  def extra_template_kwargs(self):
    return dict(
        toctree=self.toctree,
        symbols_table=self.make_symbols_tables(
            self.symbol.node.documented_childs
        ),
        source_link=github_link.get_module_link(self.symbol.value.__name__),
        **super().extra_template_kwargs,
    )

  @property
  def toctree(self) -> str:
    items = []
    for n in self.symbol.node.documented_childs:
      path = n.match.filename.relative_to(self.filename.parent)
      path = os.fspath(path)
      path = path.removesuffix('.md')
      items.append(path)
    return '\n'.join(items)


class _RootModule(_IsModule):
  recurse = True
  template_name = 'api'

  def match(self) -> bool:
    return self.symbol.parent is None

  @property
  def filename(self) -> pathlib.Path:
    # return pathlib.Path(self.symbol.name) / 'index.md'
    return pathlib.Path(self.symbol.ctx.alias) / 'index.md'

  @property
  def extra_template_kwargs(self):
    return dict(
        **super().extra_template_kwargs,
        import_statement=self.import_statement,
        all_symbols_table=self.make_symbols_tables(
            self.symbol.node.iter_documented_nodes()
        ),
    )

  @property
  def import_statement(self) -> str:
    module_name = self.symbol.ctx.module_name
    alias = self.symbol.ctx.alias

    if '.' in module_name:
      left, last_name = module_name.rsplit('.')
      left = f'from {left} '
    else:
      left = ''
      last_name = module_name

    if last_name == alias:
      stmt = f'import {last_name}'
    else:
      stmt = f'import {last_name} as {alias}'

    stmt = left + stmt
    return epy.dedent(
        f"""
        ```{{code-block}}
        {stmt}
        ```
        """
    )


class _ImplicitlyImportedModule(_IsModule):
  """Filter implicitly imported modules."""

  documented = False

  def match(self) -> bool:
    return not self.symbol.is_imported


class _ExternalModule(_IsModule):
  documented = False

  def match(self) -> bool:
    return not self.symbol.belong_to_namespace


class _ApiModule(_IsModule):
  """Modules or packages."""

  def match(self) -> bool:
    return self.symbol.belong_to_namespace

  @property
  def documented(self):
    # Only document when the parent is a package
    return _is_package(self.symbol.parent)

  @property
  def recurse(self):
    # Only recurse when the parent is a package
    return _is_package(self.symbol.parent)


class _IsValue(Match):
  match = _not(_IsModule)

  @property
  def filename(self) -> pathlib.Path:
    return (
        self.symbol.parent_symb.match.filename.parent / f'{self.symbol.name}.md'
    )


class _PrivateValue(_IsValue):
  documented = False

  def match(self) -> bool:
    return self.symbol.name.startswith('_')


class _MagicModuleAttribute(_IsValue):
  documented = False

  def match(self):
    return self.symbol.name in [
        '__name__',
        '__doc__',
        '__package__',
        '__loader__',
        '__spec__',
        '__path__',
        '__file__',
        '__cached__',
        '__builtins__',
    ]


class _FutureAnnotation(_IsValue):
  documented = False

  def match(self):
    import __future__

    return isinstance(self.symbol.value, __future__._Feature)


class _DocumentedValue(_IsValue):

  @property
  def documented(self):
    # Only document imported values when the parent is a package
    return not self.symbol.is_imported or _is_package(self.symbol.parent)


# TODO(epot): How to duplicate this with _ImportedValue ?


class _WithSourceLink(Match):

  @functools.cached_property
  def _ast_symbol(self):
    module_name = self.symbol.parent.__name__
    name = self.symbol.name

    return ast_utils.extract_last_symbol(module_name, name)

  @property
  def docstring_1line(self) -> str:
    doc = ' '.join(self._ast_symbol.docstring.split('\n'))
    if len(doc) > 83:  # Truncate
      return doc[:80] + '...'
    else:
      return doc

  @property
  def extra_template_kwargs(self):
    module_name = self.symbol.parent.__name__
    name = self.symbol.name

    source_link = github_link.get_assignement_link(module_name, name)

    return dict(
        source_link=source_link,
        source_code=self._ast_symbol.code,
        docstring=self._ast_symbol.docstring,
        **super().extra_template_kwargs,
    )


class _TypeAliasValue(_DocumentedValue, _WithSourceLink):
  icon = SymbolType.TYPING
  template_name = 'type_alias'

  def match(self):
    # TODO(epot): How to detect `Any`,...

    return (
        isinstance(self.symbol.value, typing.TypeVar)
        or typing_extensions.get_origin(self.symbol.value) is not None
    )


class _ClassValue(_WithDocstring, _DocumentedValue):
  icon = SymbolType.CLASS
  template_name = 'class'

  def match(self):
    return isinstance(self.symbol.value, type)


class _FunctionValue(_WithDocstring, _DocumentedValue):
  icon = SymbolType.FUNCTION
  template_name = 'function'

  def match(self):
    return isinstance(
        self.symbol.value,
        (
            types.FunctionType,
            types.BuiltinFunctionType,
            types.BuiltinMethodType,
            types.MethodType,
            types.MethodWrapperType,
            functools._lru_cache_wrapper,  # `@functools.wraps`
        ),
    )


class _AttributeValue(_DocumentedValue, _WithSourceLink):
  icon = SymbolType.ATTRIBUTE
  template_name = 'attribute'


def _is_package(module: types.ModuleType) -> bool:
  # Have custom attribute so standard module can behave like package.
  if hasattr(module, '__apitree__'):
    return module.__apitree__['is_package']
  return module.__name__ == module.__package__



from etils import epy


class Table:

  def __init__(self, *, header):
    self._lines = epy.Lines()
    self._lines += '| ' + ' | '.join(header) + ' |'
    self._lines += ' | '.join('---' for _ in header)

  def add_row(self, *row: str):
    self._lines += ' | '.join(row)

  def make(self) -> str:
    return self._lines.join()



import functools
import importlib
import inspect
import pathlib
import subprocess


@functools.cache
def module_path(module_name: str) -> pathlib.Path:
  """Get absolute path of a module."""
  root_module = importlib.import_module(module_name)
  root_module_path = inspect.getsourcefile(root_module)
  root_module_path = pathlib.Path(root_module_path)
  return root_module_path


# TODO(epot): OO API to allow multiple context with various origins
def repo_path() -> pathlib.Path:
  """Absolute path of the repository."""
  out = subprocess.run(
      'git rev-parse --show-toplevel',
      shell=True,
      capture_output=True,
      text=True,
  )
  path = out.stdout.strip()
  return pathlib.Path(path)


def abs_path(module_name: str) -> pathlib.Path:
  filepath = inspect.getsourcefile(importlib.import_module(module_name))
  filepath = pathlib.Path(filepath)
  return filepath


@functools.cache
def module_lines(module_name: str) -> list[str]:
  filepath = abs_path(module_name)
  return filepath.read_text().split('\n')


def belong_to_repo(module_name: str) -> bool:
  return repo_relative_path(module_name) is not None


def repo_relative_path(module_name: str) -> pathlib.Path:
  """."""
  filepath = abs_path(module_name)
  try:
    return filepath.relative_to(repo_path())
  except ValueError:
    return None



"""Fix bad ```python md formatting."""

from sphinx.application import Sphinx

from apitree import debug_utils


def _preprocess_docstring(app, what, name, obj, options, lines):
  # Modify each line of the docstring
  is_block = False
  new_lines = []
  for line in lines:
    if line == '```python':
      assert not is_block
      new_lines.append('.. code-block::')
      new_lines.append('')
      is_block = True
    elif line == '```':
      if is_block:
        new_lines.append('')
      else:
        new_lines.append('.. code-block::')
        new_lines.append('')
      is_block = not is_block
    else:
      if is_block:
        line = f'  {line}'
      new_lines.append(line)
  lines[:] = new_lines


def setup(app: Sphinx):
  # Fix bad ```python md formatting
  app.connect(
      'autodoc-process-docstring',
      debug_utils.print_error()(_preprocess_docstring),
  )
